AllaraBossCore = LibStub("AceAddon-3.0"):NewAddon("AllaraBossCore", "AceConsole-3.0", "AceEvent-3.0", "LibShefkiTimer-1.0", "AceComm-3.0", "AceSerializer-3.0")
AllaraBossCore.MINOR_VERSION = tonumber(("$Revision: 2 $"):match("%d+"))
local L = LibStub("AceLocale-3.0"):GetLocale("AllaraBossCore")
local ABC = AllaraBossCore
local zonenames = LibStub("LibBabble-Zone-3.0"):GetLookupTable()
local maps = LibStub("LibMapData-1.0")
--



--[[
      VARIABLES ----------------------------------------------------------------------------------------------------------------
  ]]

local POLLING_FREQUENCY = 1.0
local possibleWipe = false
local targetList = {}
local pullChecks = 0
local commPrefix = "AllaraBossCore"
local startingFight = false

ABC.isEnabled = false
ABC.inCombat = false
ABC.ver = {}
ABC.raid = {}
ABC.enc = {}
ABC.zones = {}
ABC.fight = nil
ABC.testZone = nil
ABC.forceHeroic = nil
ABC.testing = false
ABC.forcedTargetNpc = nil
ABC.forcedTargetName = nil
ABC.modules = {}

StaticPopupDialogs["ALLARABOSSCORE_UPGRADE_NOTICE"] = {
	text = "",
	button1 = ACCEPT,
	hideOnEscape = 1,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
}



--[[
      UPVALUES -----------------------------------------------------------------------------------------------------------------
  ]]
  
local _G = getfenv(0)
local pairs = _G.pairs
local strlower = _G.strlower
local strupper = _G.strupper
local format = _G.format

local UnitAffectingCombat = _G.UnitAffectingCombat
local UnitGUID = _G.UnitGUID
local UnitName = _G.UnitName
local UnitInRaid = _G.UnitInRaid
local UnitIsDeadOrGhost = _G.UnitIsDeadOrGhost



--[[
      INITIALIZATION -----------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnInitialize()
	-- Register for slash commands
	self:RegisterChatCommand("allarabosscore", "ChatCommand")
end


function ABC:OnEnable()
	-- Scan for add-ons
	self.modules = {}
	for i = 1, GetNumAddOns() do
		if GetAddOnMetadata(i, "X-AllaraBossCore-LoadZone") then
			table.insert(self.modules, {
				addon = GetAddOnInfo(i),
				zone = {strsplit(",", GetAddOnMetadata(i, "X-AllaraBossCore-LoadZone") or "")},
				minrev = GetAddOnMetadata(i, "X-AllaraBossCore-MinimumRevision") or 1
			})
		end
	end
	local noticed = false
	for k, v in pairs(self.modules) do
		if v.minrev and (tonumber(v.minrev) or 1) > ABC.MINOR_VERSION then
			self:Print(format(L["|cffffff20%s|r requires Allara Boss Core revision |cffffff20%d|r or higher. It will not function until you update."], v.addon, v.minrev))
			if not noticed then
				StaticPopupDialogs["ALLARABOSSCORE_UPGRADE_NOTICE"].text = format(L["|cffffff20%s|r requires Allara Boss Core revision |cffffff20%d|r or higher. It will not function until you update."].."\n\n"..L["Please update by navigating to:"].."\n\n%s",
					v.addon, v.minrev, "http://www.wowace.com/addons/allara-boss-core/")
				StaticPopup_Show("ALLARABOSSCORE_UPGRADE_NOTICE")
				noticed = true
			end
			self.modules[k] = nil
		end
	end

	-- Register for events
	self:UnregisterAllEvents()
	self:RegisterEvent("RAID_ROSTER_UPDATE", "OnEvent")
	self:RegisterEvent("PLAYER_REGEN_DISABLED", "OnEvent")
	self:RegisterEvent("PLAYER_REGEN_ENABLED", "OnEvent")
	self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_YELL", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_EMOTE", "OnEvent")
	self:RegisterEvent("CHAT_MSG_RAID_BOSS_EMOTE", "OnEvent")
	self:RegisterEvent("CHAT_MSG_MONSTER_SAY", "OnEvent")

	-- Register communication
	self:UnregisterAllComm()
 self:RegisterComm(commPrefix, "OnCommReceived")

	-- Initialize
	self:LoadZone()
	self:UpdateRaid()
	
	-- Turn us on
	ABC.isEnabled = true
	
	-- Ask the raid if we should be in a fight right now
	if UnitInRaid("player") then
		self:SendCommMessage(commPrefix, self:Serialize({ c = "FIGHT_REQUEST" }), "RAID")
	end

	--[===[@debug@
	self:Print("Allara Boss Core loaded (running in debug mode)")
	--@end-debug@]===]
end


function ABC:OnDisable()
	self:EndFight()

	-- Cancel all timers
	self:CancelAllTimers()

	-- Turn us off
	ABC.isEnabled = false
	
	--[===[@debug@
	self:Print(L["Disabled"])
	--@end-debug@]===]
end


function ABC:ChatCommand(input)
	if strlower(input) == "ver" then
		self:VersionCheck()
	end
end


function ABC:VersionCheck()
	if UnitInRaid("player") then
		ver = {}
		self:SendCommMessage(commPrefix, self:Serialize({ c = "VER_REQ" }), "RAID")
		self:ScheduleTimer("OnVersionCheckFinished", 5)
	end
end

function ABC:OnVersionCheckFinished()
	if UnitInRaid("player") then
		for k, v in pairs(ABC.raid) do
			if not ver[k] and UnitInRaid(k) and v.online then
				self:Print(format(L["%s is not using Allara Boss Core"], k))
			end
		end
		local count = 0
		for k, v in pairs(ver) do count = count + 1 end
		local numNot = GetNumRaidMembers() - count
		if numNot < 0 then numNot = 0 end
		self:Print(format(L["%d player(s) are running Allara Boss Core. %d player(s) are not."], count, numNot))
	end
end



--[[
      EVENTS -------------------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnEvent(event, ...)

	if event == "RAID_ROSTER_UPDATE" then
		self:UpdateRaid()
		
	elseif event == "ZONE_CHANGED_NEW_AREA" then
		self:LoadZone()
						
	elseif event == "PLAYER_REGEN_DISABLED" then
		if UnitInRaid("player") then
			self.inCombat = true
			self:CheckForPull()
		end
			
	elseif event == "PLAYER_REGEN_ENABLED" then
		self.inCombat = false
		self:UpdateRaid()
		
	elseif event == "CHAT_MSG_MONSTER_YELL" then
		local msg = select(1, ...)
		if ABC.fight and ABC.fight.onYell then ABC.fight.onYell(ABC.fight, event, ...) end
		self:OnMonsterMessage("yell", msg)
	
	elseif event == "CHAT_MSG_MONSTER_EMOTE" then
		local msg = select(1, ...)
		if ABC.fight and ABC.fight.onYell then ABC.fight.onYell(ABC.fight, event, ...) end
		self:OnMonsterMessage("emote", msg)
	
	elseif event == "CHAT_MSG_RAID_BOSS_EMOTE" then
		local msg = select(1, ...)
		if ABC.fight and ABC.fight.onYell then ABC.fight.onYell(ABC.fight, event, ...) end
		self:OnMonsterMessage("emote", msg)
	
	elseif event == "CHAT_MSG_MONSTER_SAY" then
		local msg = select(1, ...)
		if ABC.fight and ABC.fight.onYell then ABC.fight.onYell(ABC.fight, event, ...) end
		self:OnMonsterMessage("say", msg)
					
 end
end


function ABC:OnCombatEvent(event, timestamp, eventType, srcGuid, srcName, srcFlags, dstGuid, dstName, dstFlags, ...)
	-- local a = select(1, ...)
	-- self:Print(format("%s %s %s %s", eventType, srcName or "nil", dstName or "nil", a or "nil"))
	if ABC.fight and ABC.isEnabled then

		if ABC.fight.combatEvents and (ABC.fight.combatEvents.ALL or ABC.fight.combatEvents[eventType]) then
			if ABC.fight.onCombat then
				ABC.fight.onCombat(ABC.fight, event, timestamp, eventType, srcGuid, srcName, srcFlags, dstGuid, dstName, dstFlags, ...)
			end
		end
		
		if eventType == "UNIT_DIED" or eventType == "UNIT_DESTROYED" then
			if bit.band(dstGuid:sub(1, 5), 0x00F) == 3 or bit.band(dstGuid:sub(1, 5), 0x00F) == 5 then
				self:OnMobKill(tonumber(dstGuid:sub(9, 12), 16))
			end
		end

	end
end


function ABC:OnCommReceived(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
		--[===[@debug@
  self:Print("[OnCommReceived] Failed to deserialize message")
		--@end-debug@]===]
  return
 else
		if o.c then
		
			if o.c == "VER" then
				if o.v then
					ver[sender] = o.v
					self:Print(format(L["%s is using Allara Boss Core r%s"], sender, o.v))
				end

			elseif o.c == "VER_REQ" then
				self:SendCommMessage(commPrefix, self:Serialize({ c = "VER", v = self.MINOR_VERSION }), "WHISPER", sender)

			elseif o.c == "FIGHT_REQUEST" and sender ~= UnitName("player") then
				if ABC.fight then
					self:SendCommMessage(commPrefix, self:Serialize({ c = "FIGHT_RESPONSE", f = ABC.fight.enable.fight }), "WHISPER", sender)
				else
				end
				
			elseif o.c == "FIGHT_RESPONSE" and sender ~= UnitName("player") then
				if not ABC.fight and not startingFight then
					if ABC.enc[o.f] then
						startingFight = true
						self:LoadZone(true)
						self:ScheduleTimer(function()
							self:StartFight(ABC.enc[o.f])
							startingFight = false
						end, 3.0)
					end
				end
								
			end
			
		end
 end
end



--[[
      TARGET DETECTION ---------------------------------------------------------------------------------------------------------
  ]]
  
do

	--[[
	
		GetBossTargetChange (typically useful for SPELL_CAST_START events during
		really picky mechanics):
		
			If the boss is your target or focus, start watching for UNIT_TARGET
			events. If one is found, use the unit's target at that point. If one
			is not found within 0.3 seconds, use the target at 0.3 seconds.
			
			If the boss isn't your target or focus, scan the raid to find it.
			Save it's target's GUID. Scan every 0.05 seconds for 0.3 seconds and see if
			the boss changes targets during that time. If so, use the new target.
			If not, use the target at 0.3 seconds.
			
		GetBossTarget (typically useful for SPELL_CAST_COMPLETE events):
		
		 After a configurable delay, scans your target, focus, and the entire
		 raid to find the boss. Once found, uses the boss target.
			
	]]

	local isScanning = false
	local lastGuid
	local interested_unit
	local tries = 0
	local tryTimer, cancelTimer
	local FAILSAFE_TIME = 0.3
	local TRY_REPEAT_TIME = 0.05
	local MAX_TRIES = 6
	local fn, cid
	local inDoTarget = false
	
	
	local function deinit()
		if tryTimer then ABC:CancelTimer(tryTimer, true) end
		if cancelTimer then ABC:CancelTimer(cancelTimer, true) end
		isScanning = false
		lastGuid = nil
		interested_unit = nil
		fn = nil
		cid = nil
		tryTimer = nil
		cancelTimer = nil
	end
	
	local function doTarget(unit)
		if not inDoTarget then
			inDoTarget = true
			local u = unit.."target"
			if UnitExists(u) then
				if fn and type(fn) == "function" then
					local n = UnitName(u)
					if n then
						--[===[@debug@
						ABC:Print(format("Using boss target for unit '%s': '%s'", u, n))
						--@end-debug@]===]
						fn(n)
					end
				end
			end
			deinit()
			inDoTarget = false
		end
	end


	function ABC:OnTarget(event, unit) -- Fires for UNIT_TARGET events
		if isScanning and unit == interested_unit then
			--[===[@debug@
			ABC:Print(format("UNIT_TARGET fired for unit: '%s'", unit))
			--@end-debug@]===]
			doTarget(unit)
		end
	end
	
	
	-- failsafe doesn't get scheduled for a normal repeated scan
	local function failsafe()
		--[===[@debug@
		ABC:Print("FAILSAFE fired")
		--@end-debug@]===]
		if interested_unit then
			doTarget(interested_unit)
		else
			local name, unit, guid = scan(cid)
			if unit then
				doTarget(unit)
			else
				deinit()
			end
		end
	end
	
	
	local t = {"target", "targettarget", "focus", "focustarget", "mouseover", "mouseovertarget"}
	for i = 1, 4 do t[#t+1] = format("boss%d", i) end
	for i = 1, 4 do t[#t+1] = format("party%dtarget", i) end
	for i = 1, 40 do t[#t+1] = format("raid%dtarget", i) end
	local function findBoss(id)
		local idType = type(id)
		for i, unit in next, t do
			if UnitExists(unit) and not UnitIsPlayer(unit) then
				local unitId = UnitGUID(unit)
				if idType == "number" then unitId = tonumber(unitId:sub(-12, -7), 16) end
				if unitId == id then return unit end
			end
		end
	end
	

	local function scan(id)
		local unit = findBoss(id)
		if not unit then return end
		if UnitExists(unit.."target") then return UnitName(unit), unit, UnitGUID(unit) end
	end
	
	
	local function try()
		tries = tries + 1
		local name, unit, guid = scan(cid)
		if unit and guid and guid ~= lastGuid then
			doTarget(unit)
			return
		end
		
		if tries == MAX_TRIES then
			if unit then
				doTarget(unit)
			else
				deinit()
			end
		end
	end


	function ABC:GetBossTargetChange(_cid, _fn)
		deinit()
		inDoTarget = false
		
		fn = _fn
		cid = _cid

		--[===[@debug@
		if ABC.forcedTargetNpc == cid and ABC.forcedTargetName and fn then
			fn(ABC.forcedTargetName)
			return
		end
		--@end-debug@]===]
		
		if self:GetUnitCreatureId("target") == cid then
			interested_unit = "target"
			isScanning = true
			cancelTimer = self:ScheduleTimer(failsafe, FAILSAFE_TIME)
		elseif self:GetUnitCreatureId("focus") == cid then
			interested_unit = "focus"
			isScanning = true
			cancelTimer = self:ScheduleTimer(failsafe, FAILSAFE_TIME)
		else
			interested_unit = nil
			local name, unit, guid = scan(cid)
			lastGuid = guid
			self:ScheduleRepeatingTimer(try, TRY_REPEAT_TIME)
			tries = 0
		end
	end
	
	function ABC:GetBossTarget(_cid, _fn, _delay)
		--[===[@debug@
		if ABC.forcedTargetNpc == _cid and ABC.forcedTargetName and _fn then
			_fn(ABC.forcedTargetName)
			return
		end
		--@end-debug@]===]
		self:ScheduleTimer(function()
			local name, unit, guid = scan(_cid)
			if name and type(_fn) == "function" then
				_fn(name)
			end
		end, _delay or 0.1)
	end
	
end



--[[
      PLAYER TYPE DETECTION ----------------------------------------------------------------------------------------------------
  ]]
  
local function getTalentpointsSpent(spellID)
	local spellName = GetSpellInfo(spellID)
	for tabIndex=1, GetNumTalentTabs() do
		for talentID=1, GetNumTalents(tabIndex) do
			local name, _, _, _, spent = GetTalentInfo(tabIndex, talentID)
			if(name == spellName) then
				return spent
			end
		end
	end
	return 0
end

function ABC:IsMelee()
	return select(2, UnitClass("player")) == "ROGUE"
		or select(2, UnitClass("player")) == "WARRIOR"
		or select(2, UnitClass("player")) == "DEATHKNIGHT"
		or (select(2, UnitClass("player")) == "PALADIN" and select(3, GetTalentTabInfo(1)) < 51)
		or (select(2, UnitClass("player")) == "SHAMAN" and select(3, GetTalentTabInfo(2)) >= 51)
		or (select(2, UnitClass("player")) == "DRUID" and select(3, GetTalentTabInfo(2)) >= 51)
end

function ABC:IsRanged()
	return select(2, UnitClass("player")) == "MAGE"
		or select(2, UnitClass("player")) == "HUNTER"
		or select(2, UnitClass("player")) == "WARLOCK"
		or select(2, UnitClass("player")) == "PRIEST"
		or (select(2, UnitClass("player")) == "PALADIN" and select(3, GetTalentTabInfo(1)) >= 51)
		or (select(2, UnitClass("player")) == "SHAMAN" and select(3, GetTalentTabInfo(2)) < 51)
		or (select(2, UnitClass("player")) == "DRUID" and select(3, GetTalentTabInfo(2)) < 51)
end

function ABC:IsPhysical()
	return self:IsMelee() or select(2, UnitClass("player")) == "HUNTER"
end

function ABC:CanRemoveEnrage()
	return select(2, UnitClass("player")) == "HUNTER" or select(2, UnitClass("player")) == "ROGUE"
end

local function IsDeathKnightTank()
	local tankTalents = (getTalentpointsSpent(16271) >= 5 and 1 or 0) +		-- Anticipation
	                    (getTalentpointsSpent(49042) >= 5 and 1 or 0) +		-- Toughness
																					(getTalentpointsSpent(55225) >= 5 and 1 or 0)	  	-- Blade Barrier
	return tankTalents >= 2
end

local function IsDruidTank()
	local tankTalents = (getTalentpointsSpent(57881) >= 2 and 1 or 0) +		-- Natural Reaction
	                    (getTalentpointsSpent(16929) >= 3 and 1 or 0) +		-- Thick Hide
																					(getTalentpointsSpent(61336) >= 1 and 1 or 0) +		-- Survival Instincts
																					(getTalentpointsSpent(57877) >= 3 and 1 or 0)		  -- Protector of the Pack
	return tankTalents >= 3
end

function ABC:IsTank()
	return (select(2, UnitClass("player")) == "WARRIOR" and select(3, GetTalentTabInfo(3)) >= 51)
		or (select(2, UnitClass("player")) == "DEATHKNIGHT" and IsDeathKnightTank())
		or (select(2, UnitClass("player")) == "PALADIN" and select(3, GetTalentTabInfo(2)) >= 51)
		or (select(2, UnitClass("player")) == "DRUID" and select(3, GetTalentTabInfo(2)) >= 51 and IsDruidTank())
end

function ABC:IsHealer()
	return (select(2, UnitClass("player")) == "PALADIN" and select(3, GetTalentTabInfo(1)) >= 51)
		or (select(2, UnitClass("player")) == "SHAMAN" and select(3, GetTalentTabInfo(3)) >= 51)
		or (select(2, UnitClass("player")) == "DRUID" and select(3, GetTalentTabInfo(3)) >= 51)
		or (select(2, UnitClass("player")) == "PRIEST" and select(3, GetTalentTabInfo(3)) < 51)
end



--[[
      PULL DETECTION -----------------------------------------------------------------------------------------------------------
  ]]
  
-- Combat-based detection (some raid member is targeting the boss, and the boss is affecting combat)
function ABC:BuildTargetList()
	local uId = "raid"
	for i = 0, GetNumRaidMembers() do
		local id = (i == 0 and "target") or uId..i.."target"
		local guid = UnitGUID(id)
		if guid and (bit.band(guid:sub(1, 5), 0x00F) == 3 or bit.band(guid:sub(1, 5), 0x00F) == 5) then
			local cId = tonumber(guid:sub(9, 12), 16)
			targetList[cId] = id
		end
	end
end

function ABC:ClearTargetList()
	table.wipe(targetList)
end

function ABC:DoDelayedCheckForPull(mob, fight)
	if not fight.inCombat then
		ABC:BuildTargetList()
		pullChecks = pullChecks + 1
		if targetList[mob] and UnitAffectingCombat(targetList[mob]) then
			--[===[@debug@
			self:Print("Pull detected")
			--@end-debug@]===]
			ABC:StartFight(fight, 3)
		elseif targetList[mob] and pullChecks < 10 then
			ABC:ScheduleTimer(function() ABC:DoDelayedCheckForPull(mob, fight) end, 3 * pullChecks)
		end
		ABC:ClearTargetList()
	end
end

function ABC:DoCheckForPull(mob, fight)
	local uId = targetList[mob]
	if uId and UnitAffectingCombat(uId) then
		--[===[@debug@
		self:Print("Pull detected")
		--@end-debug@]===]
		ABC:StartFight(fight, 0)
		return true
	elseif uId then
		ABC:ScheduleTimer(function() ABC:DoDelayedCheckForPull(mob, fight) end, 3)
	end
end

function ABC:CheckForPull()
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		pullChecks = 0
		ABC:BuildTargetList()
		for k, v in pairs(ABC.enc) do
			if v.pull and v.pull.type == "combat" and v.pull.npcs then
				for kk, vv in pairs(v.pull.npcs) do
					if ABC:DoCheckForPull(vv, v) then break end
				end
			end
		end
		ABC:ClearTargetList()
	end
end


-- Yell-based detection (used for some vehicle fights, etc.)
function ABC:OnMonsterMessage(type, msg)

	-- Pull detection
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		for k, v in pairs(ABC.enc) do
			if v.pull and v.pull.type == "yell" and v.pull.yells then
				for kk, vv in pairs(v.pull.yells) do
					if vv == msg then
						--[===[@debug@
						self:Print("Pull detected")
						--@end-debug@]===]
						ABC:StartFight(v, 0)
					end
				end
			end
		end
	end

	-- Defeat detection
	if ABC.zones and (ABC.zones[GetRealZoneText()] or (ABC.testZone and ABC.zones[ABC.testZone])) then
		for k, v in pairs(ABC.enc) do
			if v.defeat and v.defeat.type == "yell" and v.defeat.yells then
				for kk, vv in pairs(v.defeat.yells) do
					if vv == msg and v.inCombat then
						--[===[@debug@
						self:Print("Defeat detected")
						--@end-debug@]===]
						ABC:EndFight()
					end
				end
			end
		end
	end

end
	


--[[
      WIPE DETECTION -----------------------------------------------------------------------------------------------------------
  ]]

function ABC:CheckForWipe()
	local wipe = true
	local uId = ((GetNumRaidMembers() == 0) and "party") or "raid"
	for i = 0, math.max(GetNumRaidMembers(), GetNumPartyMembers()) do
		local id = (i == 0 and "player") or uId..i
		if UnitAffectingCombat(id) and not UnitIsDeadOrGhost(id) then
			wipe = false
			break
		end
	end
	return wipe
end



--[[
      DEFEAT DETECTION ---------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnMobKill(id)
	if ABC.fight and ABC.fight.defeat and ABC.fight.defeat.type == "death" then
		if not ABC.fight.dead then ABC.fight.dead = {} end
		if ABC.fight.defeat.npcs then
			for k, v in pairs(ABC.fight.defeat.npcs) do
				if id == v then
					ABC.fight.dead[v] = true
				end
			end
			local allDead = true
			for k, v in pairs(ABC.fight.defeat.npcs) do
				if not ABC.fight.dead[v] then
					allDead = false
					break
				end
			end
			if allDead then
				if ABC.fight.onDefeat then ABC.fight.onDefeat(ABC.fight) end
				--[===[@debug@
				self:Print("Defeat detected")
				--@end-debug@]===]
				self:EndFight()
			end
		end
	end
end



--[[
      FIGHT LOGIC --------------------------------------------------------------------------------------------------------------
  ]]

function ABC:OnTimer()
	self:UpdateRaid()
	if ABC.fight and ABC.fight.onHeartbeat and type(ABC.fight.onHeartbeat) == "function" then ABC.fight.onHeartbeat(ABC.fight) end
end


function ABC:UpdateRaid()
	if UnitInRaid("player") then
	
		for k, v in pairs(ABC.raid) do
			ABC.raid[k].valid = nil
		end
	
		-- Gather data about the raid
		local numRaid = GetNumRaidMembers()
		for i = 1, numRaid do
			local playerName, rank, subgroup, level, class, fileName, zone, online, isDead, role, isML = GetRaidRosterInfo(i)
			if playerName then
				if not ABC.raid[playerName] then ABC.raid[playerName] = {} end
				if not online and ABC.raid[playerName].online and ABC.fight and ABC.fight.onDisconnect and type(ABC.fight.onDisconnect) == "function" then ABC.fight.onDisconnect(ABC.fight, playerName) end
				ABC.raid[playerName].rank = rank
				ABC.raid[playerName].subgroup = subgroup
				ABC.raid[playerName].level = level
				ABC.raid[playerName].class = class
				ABC.raid[playerName].zone = zone
				ABC.raid[playerName].online = online
				ABC.raid[playerName].isDead = isDead
				ABC.raid[playerName].role = role
				ABC.raid[playerName].isML = isML
				ABC.raid[playerName].valid = true
			end -- valid player at that index
		end -- player loop
		
		-- Delete players no longer in the raid
		for k, v in pairs(ABC.raid) do
			if not v.valid then
				ABC.raid[k] = nil
			end
		end
		
		-- Check for a wipe
		if ABC.isEnabled and ABC.fight and self:CheckForWipe() and not possibleWipe then
			possibleWipe = true
			local time = 5
			if ABC.fight.wipeTime then time = ABC.fight.wipeTime end
			--[===[@debug@
			self:Print(format("Possible wipe detected. Checking again in %d seconds.", time))
			--@end-debug@]===]
			self:ScheduleTimer(function()
				if possibleWipe then
					if ABC:CheckForWipe() then
						possibleWipe = false
						if ABC.fight.onWipe then ABC.fight.onWipe(ABC.fight) end
						--[===[@debug@
						self:Print("Wipe detected")
						--@end-debug@]===]
						ABC:EndFight()
					end
				end
			end, time)
		end
		
	else
		ABC.raid = {}
	end -- we're in a raid
end


function ABC:CancelPossibleWipe()
	possibleWipe = false
end


function ABC:LoadZone(zone, now)
	if zone then ABC.testZone = zone end
	for k, v in pairs(self.modules) do
		for kk, vv in pairs(v.zone) do
			--self:Print(format("real zone: %s; checking addon: %s; addon wants zone: %s; babble zone is: %s", GetRealZoneText(), v.addon, vv, zonenames[vv]))
			if (GetRealZoneText() == zonenames[vv] or (zone and zone == zonenames[vv])) and v.addon and not IsAddOnLoaded(v.addon) then
				local inInstance, instanceType = IsInInstance()
				if (inInstance and instanceType == "raid") or zone then
					if now then
						LoadAddOn(v.addon)
					else
						self:ScheduleTimer(function() LoadAddOn(v.addon) end, 3)
					end
				end
			end
		end
	end
end


function ABC:StartFight(fight, startDelay, silent)
	if type(fight) == "string" then
		if ABC.enc[fight] then
			fight = ABC.enc[fight]
		else
			error(format("Unable to locate fight: %s", fight))
			return
		end
	end
	
	if not fight then
		error("No fight specified")
		return
	end

	if fight.enabled ~= false and ABC.isEnabled then
		-- Register for events
		self:RegisterEvent("UNIT_TARGET", "OnTarget")
		self:RegisterEvent("COMBAT_LOG_EVENT_UNFILTERED", "OnCombatEvent")

		-- Initialize fight
		possibleWipe = false
		fight.inCombat = true
		fight.startDelay = startDelay
		fight.dead = {}
		ABC.fight = fight
		if not ABC.testing then
			self:CancelAllTimers()
			self:ScheduleRepeatingTimer("OnTimer", POLLING_FREQUENCY)
		end
		
		--[===[@debug@
		if not silent then
			local heroic = ""
			if ABC:IsHeroicRaid() then heroic = " (heroic)" end
			self:Print(format("Initiating fight: %s"..heroic, fight.name or "Name not specified in encounter definition"))
		end
		--@end-debug@]===]
		
		if fight.onStart then fight.onStart(fight) end
		
		return true -- indicate success to callers
	end
end


function ABC:EndFight(silent)
	self:CancelAllTimers()
	self:UnregisterEvent("UNIT_TARGET")
	self:UnregisterEvent("COMBAT_LOG_EVENT_UNFILTERED")
	if ABC.fight then
		ABC.fight.inCombat = false
		--[===[@debug@
		if not silent then self:Print(format("Ending fight: %s", ABC.fight.name)) end
		--@end-debug@]===]
		if ABC.fight.onEnd then ABC.fight.onEnd(ABC.fight) end
		ABC.fight = nil
	end
end



--[[
      API ----------------------------------------------------------------------------------------------------------------------
  ]]

function ABC:RegisterEncounters(encounters)
	for k, v in pairs(encounters) do
		self:RegisterEncounter(k, v)
	end
end


function ABC:RegisterEncounter(name, definition)
	if self:ValidateEncounter(name, definition) then
		ABC.enc[name] = definition
		if definition.zone then ABC.zones[zonenames[definition.zone]] = true end
		ABC.enc[name].inCombat = false
	end
end


function ABC:ValidateEncounter(name, definition)
	return true
end


function ABC:IsHeroicRaid()
	if ABC.forceHeroic ~= nil then return ABC.forceHeroic end
	local diff = GetInstanceDifficulty()
	local name, type, difficultyIndex, difficultyName, maxPlayers, dynamicDifficulty, isDynamic = GetInstanceInfo()
	if isDynamic then
		if dynamicDifficulty == 1 then return true else return false end
	end
	if diff == 3 or diff == 4 then return true else return false end
end


function ABC:Sync(o)
	self:SendCommMessage(commPrefix, self:Serialize(o), "RAID")
end


function ABC:GetUnitCreatureId(unit)
	if unit then
		local guid = UnitGUID(unit)
		if guid then
			return (guid and tonumber(guid:sub(9, 12), 16)) or 0
		end
	end
end


function ABC:GetCIDFromGUID(guid)
	if guid then
		return (guid and tonumber(guid:sub(9, 12), 16)) or 0
	end
end
